
{$A+}   { Align data }
{$B-}   { Boolean evaluation }
{$E+}   { 80x87 emulator }
{$F-}   { Force FAR calls }
{$G+}   { 80286 code }
{$I-}   { I/O checking }
{$K-}   { Smart Callbacks }
{$N-}   { 80x87 code }
{$O-}   { Overlays allowed }
{$P-}   { Open parameters }
{$T-}   { Typed pointers }
{$V-}   { String VAR checking }
{$W-}   { Windows stack frame for real mode }
{$X+}   { Extended syntax }

{$IFDEF DEBUG}
    {$D+}   { Debug information }
    {$L+}   { Local symbols }
    {$Q+}   { Overflow checking }
    {$R+}   { Range checking }
    {$S+}   { Stack checking }
    {$Y+}   { Symbol reference information }
{$ELSE}
    {$D-}   { Debug information }
    {$L-}   { Local symbols }
    {$Q-}   { Overflow checking }
    {$R-}   { Range checking }
    {$S-}   { Stack checking }
    {$Y-}   { Symbol reference information }
{$ENDIF}

{$C Fixed Preload Permanent} { DLL must have these code segment attributes }

{$M 1024,0}

LIBRARY PatchMB;

USES WinTypes, WinProcs, ToolHelp;

CONST
    cCopy = 5;

VAR
    { A "JMP FAR [ADDRESS]" instruction takes 5 bytes }
    abyOrgJmp : ARRAY [0..cCopy - 1] OF Byte;
    pToJmp, pOrgExitProc : Pointer;
    selPatch, selMessageBeep, wOfsMessageBeep, wOfsJump : Word;
    bDebugVersion, bPatched : Boolean;

(* ---- *)

PROCEDURE NewMessageBeep (wType: Word); EXPORT;
{ The replacement for WinProcs.MessageBeep. It has to take the same
  parameters and if it's a FUNCTION has TO return a variable OF the same
  TYPE. Must be marked as EXPORT! }

BEGIN
    IF (NOT bPatched) THEN Exit; { Avoid disaster }

    IF (wType <> 0) THEN
    BEGIN { Call original MessageBeep procedure if wType <> 0 }
        { Restore patched code }
        MemoryWrite (selMessageBeep, wOfsMessageBeep, @abyOrgJmp[0], cCopy);

        IF (bDebugVersion) THEN
            OutPutDebugString (#13#10'MessageBeep!');

        { Lets call the original MessageBeep function }
        WinProcs.MessageBeep (wType);

        { Patch the MessageBeep procedure again to point at NewMessageBeep }
        MemoryWrite (selMessageBeep, wOfsMessageBeep, pToJmp, cCopy);
    END; { if }
END; { NewMessageBeep }

(* ---- *)

PROCEDURE Patch; EXPORT;
{ Patch original MessageBeep procedure to call NewMessageBeep }

BEGIN
    IF (bPatched) THEN Exit; { Don't DO it twice }

    { Get selector for Patch procedure }
    selPatch := AllocSelector (HiWord (LongInt (@Patch)));

    ASM
            mov ax, offset @JmpToPatch { Get offset of jump instruction }
            mov wOfsJump, ax { Copy to variable }

            jmp @CarryOn { Proceed }

        { This code never gets executed, but we need to get the instruction
          for a JMP FAR NewMessageBeep to patch the original MessageBeep
          procedure. }
        @JmpToPatch:
            jmp NewMessageBeep { Jump to new procedure }

        @CarryOn:
    END; { asm }

    { Put pointer to JMP instruction together }
    pToJmp := Pointer (MakeLong (wOfsJump, selPatch));

    { Get selector of original MessageBeep procedure }
    selMessageBeep := HiWord (LongInt (@WinProcs.MessageBeep));

    { Lock selector in memory so that Windows cannot discard the segment }
    GlobalPageLock (selMessageBeep);

    { Get offset of MessageBeep funktion }
    wOfsMessageBeep := LoWord (LongInt (@WinProcs.MessageBeep));

    { Save first 5 bytes of MessageBeep }
    MemoryRead (selMessageBeep, wOfsMessageBeep, @abyOrgJmp[0], cCopy);

    { Patch the MessageBeep procedure }
    MemoryWrite (selMessageBeep, wOfsMessageBeep, pToJmp, cCopy);

    bPatched := TRUE; { Set flag }
END; { Patch }

(* ---- *)

PROCEDURE UnPatch; EXPORT;
{ Restore original MessageBeep procedure }

BEGIN
    IF (NOT bPatched) THEN Exit; { Nothing has been patched }

    { Restore patched code }
    MemoryWrite (selMessageBeep, wOfsMessageBeep, @abyOrgJmp[0], cCopy);
    GlobalPageUnlock (selMessageBeep); { Unlock memory page }

    bPatched := FALSE; { Set flag }
END; { UnPatch }

(* ---- *)

EXPORTS
    Patch index 1,
    UnPatch index 2;

(* ---- *)

PROCEDURE NewExitProc; FAR;
{ DLL is closing down }

BEGIN
    ExitProc := pOrgExitProc;

    UnPatch; { Restore MessageBeep procedure }
END; { NewExitProc }

(* ---- *)

BEGIN { PatchMB }
    pOrgExitProc := System.ExitProc;
    System.ExitProc := @NewExitProc;

    { Test for Windows' debug kernel }
    bDebugVersion := GetSystemMetrics (sm_Debug) <> 0;

    bPatched := FALSE; { Initialize flag }
END. { PatchMB }

